/*
  Copyright (C) 2000-2005 SKYRIX Software AG

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#ifndef __NGObjWeb_OWResponder_H__
#define __NGObjWeb_OWResponder_H__

#import <Foundation/NSObject.h>

@class WORequest, WOResponse, WOContext;

@protocol OWResponder < NSObject >

- (void)takeValuesFromRequest:(WORequest *)_request
  inContext:(WOContext *)_context;

- (id)invokeActionForRequest:(WORequest *)_request
  inContext:(WOContext *)_context;

- (void)appendToResponse:(WOResponse *)_response
  inContext:(WOContext *)_context;

@end

typedef void (*OWTakeValuesMethod)    (id, SEL, WORequest  *, WOContext *);
typedef id   (*OWInvokeMethod)        (id, SEL, WORequest  *, WOContext *);
typedef void (*OWAppendResponseMethod)(id, SEL, WOResponse *, WOContext *);

#endif /* __NGObjWeb_OWResponder_H__ */
