/* 
   SQLiteChannel+Model.h

   Copyright (C) 2003-2005 SKYRIX Software AG

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SQLite Adaptor Library

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef ___SQLite_ModelFetching_H___
#define ___SQLite_ModelFetching_H___

#import "SQLiteChannel.h"

@class NSArray;
@class EOModel;

@interface SQLiteChannel(ModelFetching)

- (EOModel *)describeModelWithTableNames:(NSArray *)_tableNames;
- (NSArray *)describeTableNames;

@end

#endif
