/*
**  OracleAdaptorChannelController.h
**
**  Copyright (c) 2007  Inverse groupe conseil inc. and Ludovic Marcotte
**
**  Author: Ludovic Marcotte <ludovic@inverse.ca>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef _OracleAdaptorChannelController_H
#define _OracleAdaptorChannelController_H

#import <GDLAccess/EODelegateResponse.h>
#import <GDLAccess/EOEntity.h>

@interface OracleAdaptorChannelController : NSObject

- (EODelegateResponse) adaptorChannel: (id) theChannel
                        willInsertRow: (NSMutableDictionary *) theRow
                            forEntity: (EOEntity *) theEntity;

- (EODelegateResponse) adaptorChannel: (id) theChannel
                        willUpdateRow: (NSMutableDictionary *) theRow
                 describedByQualifier: (EOSQLQualifier *) theQualifier;

@end

#endif
