/*
  Copyright (C) 2000-2005 SKYRIX Software AG

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include "common.h"

@interface WODragContainer : WODynamicElement
@end

@interface WODropContainer : WODynamicElement
@end

@interface WODragScript : WODynamicElement
@end

@interface WODropScript : WODynamicElement
@end

@implementation WODragContainer

+ (id)allocWithZone:(NSZone *)_zone {
  static Class realClass = Nil;
  if (realClass == Nil) {
    NSLog(@"DEPRECATED: %@ !", self);
    realClass = NSClassFromString(@"WEDragContainer");
    NSAssert(realClass, @"missing WEDragContainer !");
  }
  return [realClass allocWithZone:_zone];
}

@end

@implementation WODropContainer

+ (id)allocWithZone:(NSZone *)_zone {
  static Class realClass = Nil;
  if (realClass == Nil) {
    NSLog(@"DEPRECATED: %@ !", self);
    realClass = NSClassFromString(@"WEDropContainer");
    NSAssert(realClass, @"missing WEDropContainer !");
  }
  return [realClass allocWithZone:_zone];
}

@end

@implementation WODragScript

+ (id)allocWithZone:(NSZone *)_zone {
  static Class realClass = Nil;
  if (realClass == Nil) {
    NSLog(@"DEPRECATED: %@ !", self);
    realClass = NSClassFromString(@"WEDragScript");
    NSAssert(realClass, @"missing WEDropScript !");
  }
  return [realClass allocWithZone:_zone];
}

@end

@implementation WODropScript

+ (id)allocWithZone:(NSZone *)_zone {
  static Class realClass = Nil;
  if (realClass == Nil) {
    NSLog(@"DEPRECATED: %@ !", self);
    realClass = NSClassFromString(@"WEDropScript");
    NSAssert(realClass, @"missing WEDragScript !");
  }
  return [realClass allocWithZone:_zone];
}

@end
