/*
  Copyright (C) 2000-2005 SKYRIX Software AG

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

/*
  Not implemented yet.

  Sample (Client side):
    <MAP NAME="FrontPageMap0">
     <AREA SHAPE="RECT" COORDS="0, 0, 779, 70" HREF="../index.html">
     <AREA SHAPE="RECT" COORDS="632, 76, 732, 87" HREF="../stellen/index.html">
     <AREA SHAPE="RECT" COORDS="544, 75, 595, 88" HREF="../kontakt/index.html">
     <AREA SHAPE="RECT" COORDS="399, 74, 516, 87" HREF="../apotheker/index.html">
     <AREA SHAPE="RECT" COORDS="288, 74, 368, 87" HREF="../aerzte/index.html">
     <AREA SHAPE="RECT" COORDS="147, 74, 262, 87" HREF="../patienten/index.html">
</MAP>
<img src="images/banner.jpg" width="780" height="94" alt="Esparma" border="0" 
      usemap="#FrontPageMap0">
*/
