obj/SaxException.o: SaxException.m SaxException.h \
  /usr/include/GNUstep/Foundation/NSException.h \
  /usr/include/GNUstep/GNUstepBase/GSVersionMacros.h \
  /usr/include/GNUstep/Foundation/NSString.h \
  /usr/include/GNUstep/Foundation/NSObject.h \
  /usr/include/GNUstep/Foundation/NSObjCRuntime.h \
  /usr/include/GNUstep/GNUstepBase/preface.h \
  /usr/include/GNUstep/GNUstepBase/objc-gnu2next.h \
  /usr/include/GNUstep/GNUstepBase/GSConfig.h \
  /usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h \
  /usr/include/GNUstep/Foundation/NSZone.h \
  /usr/include/GNUstep/GNUstepBase/GNUstep.h \
  /usr/include/GNUstep/Foundation/NSDate.h \
  /usr/include/GNUstep/Foundation/NSRange.h

SaxException.h:

/usr/include/GNUstep/Foundation/NSException.h:

/usr/include/GNUstep/GNUstepBase/GSVersionMacros.h:

/usr/include/GNUstep/Foundation/NSString.h:

/usr/include/GNUstep/Foundation/NSObject.h:

/usr/include/GNUstep/Foundation/NSObjCRuntime.h:

/usr/include/GNUstep/GNUstepBase/preface.h:

/usr/include/GNUstep/GNUstepBase/objc-gnu2next.h:

/usr/include/GNUstep/GNUstepBase/GSConfig.h:

/usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h:

/usr/include/GNUstep/Foundation/NSZone.h:

/usr/include/GNUstep/GNUstepBase/GNUstep.h:

/usr/include/GNUstep/Foundation/NSDate.h:

/usr/include/GNUstep/Foundation/NSRange.h:
