(ultramode
-\n 
(story
-\n    
(title
-100 Mbit/s on Fibre to the home
)title
-\n    
(url
-http://slashdot.org/articles/99/06/06/1440211.shtml
)url
-\n    
(time
-1999-06-06 14:39:59
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-wouldn't-it-be-nice
)department
-\n    
(topic
-internet
)topic
-\n    
(comments
-20
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicinternet.jpg
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Gimp 1.2 Preview
)title
-\n    
(url
-http://slashdot.org/articles/99/06/06/1438246.shtml
)url
-\n    
(time
-1999-06-06 14:38:40
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-stuff-to-read
)department
-\n    
(topic
-gimp
)topic
-\n    
(comments
-12
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicgimp.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Sony's AIBO robot Sold Out
)title
-\n    
(url
-http://slashdot.org/articles/99/06/06/1432256.shtml
)url
-\n    
(time
-1999-06-06 14:32:51
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-stuff-to-see
)department
-\n    
(topic
-tech
)topic
-\n    
(comments
-10
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topictech2.jpg
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Ask Slashdot: Another Word for "Hacker"?
)title
-\n    
(url
-http://slashdot.org/askslashdot/99/06/05/1815225.shtml
)url
-\n    
(time
-1999-06-05 20:00:00
)time
-\n    
(author
-Cliff
)author
-\n    
(department
-hacker-vs-cracker
)department
-\n    
(topic
-news
)topic
-\n    
(comments
-385
)comments
-\n    
(section
-askslashdot
)section
-\n    
(image
-topicnews.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Corel Linux FAQ
)title
-\n    
(url
-http://slashdot.org/articles/99/06/05/1842218.shtml
)url
-\n    
(time
-1999-06-05 18:42:06
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-stuff-to-read
)department
-\n    
(topic
-corel
)topic
-\n    
(comments
-164
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topiccorel.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Upside downsides MP3.COM.
)title
-\n    
(url
-http://slashdot.org/articles/99/06/05/1558210.shtml
)url
-\n    
(time
-1999-06-05 15:56:45
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-stuff-to-think-about
)department
-\n    
(topic
-music
)topic
-\n    
(comments
-48
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicmusic.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-2 Terabits of Bandwidth
)title
-\n    
(url
-http://slashdot.org/articles/99/06/05/1554258.shtml
)url
-\n    
(time
-1999-06-05 15:53:43
)time
-\n    
(author
-CmdrTaco
)author
-\n    
(department
-faster-porn
)department
-\n    
(topic
-internet
)topic
-\n    
(comments
-66
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicinternet.jpg
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Suppression of cold fusion research?
)title
-\n    
(url
-http://slashdot.org/articles/99/06/04/2313200.shtml
)url
-\n    
(time
-1999-06-04 23:12:29
)time
-\n    
(author
-Hemos
)author
-\n    
(department
-possibly-probably
)department
-\n    
(topic
-science
)topic
-\n    
(comments
-217
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicscience.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-California Gov. Halts Wage Info Sale
)title
-\n    
(url
-http://slashdot.org/articles/99/06/04/235256.shtml
)url
-\n    
(time
-1999-06-04 23:05:34
)time
-\n    
(author
-Hemos
)author
-\n    
(department
-woo-hoo!
)department
-\n    
(topic
-usa
)topic
-\n    
(comments
-16
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicus.gif
)image
-\n  
)story
-\n 
(story
-\n    
(title
-Red Hat Announces IPO
)title
-\n    
(url
-http://slashdot.org/articles/99/06/04/0849207.shtml
)url
-\n    
(time
-1999-06-04 19:30:18
)time
-\n    
(author
-Justin
)author
-\n    
(department
-details-sketchy
)department
-\n    
(topic
-redhat
)topic
-\n    
(comments
-155
)comments
-\n    
(section
-articles
)section
-\n    
(image
-topicredhat.gif
)image
-\n  
)story
-\n
)ultramode
