/*
  Copyright (C) 2004-2005 SKYRIX Software AG

  This file is part of OpenGroupware.org.

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#ifndef __WEExtensions_WEStringTableManager_H__
#define __WEExtensions_WEStringTableManager_H__

#import <Foundation/NSObject.h>

/*
  WEStringTableManager
  
  This class manages the lookup and processing of string tables (that is, the
  ".strings" files containing label translations)
  
  The class is used by the WEResourceManager.
*/

@class NSString, NSArray, NSMutableDictionary;
@class WOComponent;

@interface WEStringTableManager : NSObject
{
  NSMutableDictionary *nameToTable;
  NSMutableDictionary *compLabelCache;
}

+ (NSArray *)availableTranslations;

/* string tables */

- (NSString *)labelForKey:(NSString *)_key component:(WOComponent *)_component;

- (NSString *)stringForKey:(NSString *)_key
  inTableNamed:(NSString *)_tableName
  withDefaultValue:(NSString *)_default
  languages:(NSArray *)_languages;

- (id)stringTableWithName:(NSString *)_tableName
  inFramework:(NSString *)_framework
  languages:(NSArray *)_languages;

@end


#include <NGObjWeb/WOApplication.h>

@interface WOApplication(WEStringTableManager)

- (NSString *)shareTranslationsDirectoryName;
- (NSString *)gsTranslationsDirectoryName;

@end

#endif /* __WEExtensions_WEStringTableManager_H__ */
