EXPORTS
	__objc_class_name_NGParameterMimeType;
	__objc_class_name_NGConcreteTextMimeType;
	__objc_class_name_NGConcreteTextVcardMimeType;
	__objc_class_name_NGConcreteApplicationMimeType;
	__objc_class_name_NGConcreteAppOctetMimeType;
	__objc_class_name_NGConcreteImageMimeType;
	__objc_class_name_NGConcreteAudioMimeType;
	__objc_class_name_NGConcreteVideoMimeType;
	__objc_class_name_NGConcreteMultipartMimeType;
	__objc_class_name_NGConcreteMessageMimeType;
	__objc_class_name_NGConcreteGenericMimeType;
	__objc_class_name_NGConcreteWildcardType ;
	__objc_class_name_NGMime;
	__objc_class_name_NGMimeBodyGenerator;
	__objc_class_name_NGMimeTextBodyGenerator;
	__objc_class_name_NGMimeRfc822BodyGenerator;
	__objc_class_name_NGMimeMultipartBodyGenerator;
	__objc_class_name_NGMimeBodyParser;
	__objc_class_name_NGMimeTextBodyParser;
	__objc_class_name_NGMimeRfc822BodyParser;
	__objc_class_name_NGMimeMultipartBodyParser;
	__objc_class_name_NGMimeBodyPart;
	__objc_class_name_NGMimeBodyPartParser;
	__objc_class_name_NGMimeCommandMap;
	__objc_class_name_NGMimeDataHandler;
	__objc_class_name_NGMimeObjectDataHandler;
	__objc_class_name_NGMimeException;
	__objc_class_name_NGMimeParserException;
	__objc_class_name_NGMimeFileDataSource;
	__objc_class_name_NGMimeFileTypeMap;
	__objc_class_name_NGMimeDefaultsFileTypeMap;
	__objc_class_name_NGMimeHeaderFieldGenerator;
	__objc_class_name_NGMimeRFC822DateHeaderFieldGenerator;
	__objc_class_name_NGMimeContentTypeHeaderFieldGenerator;
	__objc_class_name_NGMimeContentLengthHeaderFieldGenerator;
	__objc_class_name_NGMimeContentDispositionHeaderFieldGenerator;
	__objc_class_name_NGMimeStringHeaderFieldGenerator;
	__objc_class_name_NGMimeHeaderFieldGeneratorSet;
	__objc_class_name_NGMimeHeaderFieldParser;
	__objc_class_name_NGMimeRFC822DateHeaderFieldParser;
	__objc_class_name_NGMimeContentTypeHeaderFieldParser;
	__objc_class_name_NGMimeContentLengthHeaderFieldParser;
	__objc_class_name_NGMimeContentDispositionHeaderFieldParser;
	__objc_class_name_NGMimeStringHeaderFieldParser;
	__objc_class_name_NGMimeHeaderFieldParserSet;
	__objc_class_name_NGMimeContentDispositionHeaderField;
	__objc_class_name_NGMimeMultipartBody;
	__objc_class_name_NGMimePartGenerator;
	__objc_class_name_NGMimePartParser;
	__objc_class_name_NGMimeType;
	NGMimeTypeText;
	NGMimeTypeAudio;
	NGMimeTypeVideo;
	NGMimeTypeImage;
	NGMimeTypeApplication;
	NGMimeTypeMultipart;
	NGMimeTypeMessage;
	NGMimeParameterTextCharset;
	NGMimeContentDispositionInlineType;
	NGMimeContentDispositionAttachmentType;
	NGMimeContentDispositionFormType;
