/*
  Copyright (C) 2004 eXtrapola Srl

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include "StructuredTextParagraph.h"
#include "common.h"

@implementation StructuredTextParagraph

- (id)initWithString:(NSString *)_s {
  if ((self = [super init])) {
    self->_text = [_s retain];
  }
  return self;
}

- (void)dealloc {
  [self->_text release];
  [super dealloc];
}

/* accessors */

- (NSString *)text {
  return self->_text;
}

/* processing */

- (NSString *)textParsedWithDelegate:(id<StructuredTextRenderingDelegate>)_del
  inContext:(NSDictionary *)_ctx 
{
  self->_delegate = _del;
  [self beginParagraphInContext:_ctx];
  [self parseText:[self text] inContext:_ctx];
  [self endParagraphInContext:_ctx];
  return nil;
}

@end /* StructuredTextParagraph */
