EXPORTS
	NGPollDescriptor;
	NGGetDescriptorFlags;
	NGSetDescriptorFlags;
	NGAddDescriptorFlag;
	NGDescriptorSend;
	NGDescriptorRecv;
	NGDescriptorIsAtty;
	NGDescriptorGetTtyName;
	__objc_class_name_NGBase64Stream;
	__objc_class_name_NGBufferedStream;
	__objc_class_name_NGByteBuffer;
	__objc_class_name_NGByteCountStream;
	__objc_class_name_NGCTextStream;
	__objc_class_name_NGCharBuffer;
	__objc_class_name_NGConcreteStreamFileHandle;
	__objc_class_name_NGCouldNotCloseStreamException;
	__objc_class_name_NGCouldNotOpenStreamException;
	__objc_class_name_NGDataStream;
	__objc_class_name_NGEndOfStreamException;
	__objc_class_name_NGFileStream;
	__objc_class_name_NGFilterStream;
	__objc_class_name_NGFilterTextStream;
	__objc_class_name_NGIOAccessException;
	__objc_class_name_NGIOException;
	__objc_class_name_NGIOSearchAccessException;
	__objc_class_name_NGLockingStream;
	__objc_class_name_NGReadOnlyStreamException;
	__objc_class_name_NGStream;
	__objc_class_name_NGStreamCoder;
	__objc_class_name_NGStreamErrorException;
	__objc_class_name_NGStreamException;
	__objc_class_name_NGStreamModeException;
	__objc_class_name_NGStreamNotOpenException;
	__objc_class_name_NGStreamPipe;
	__objc_class_name_NGStreamPipe;
	__objc_class_name_NGStreamReadErrorException;
	__objc_class_name_NGStreamSeekErrorException;
	__objc_class_name_NGStreamWriteErrorException;
	__objc_class_name_NGStreams;
	__objc_class_name_NGStringTextStream;
	__objc_class_name_NGTextStream;
	__objc_class_name_NGUnknownStreamModeException;
	__objc_class_name_NGWriteOnlyStreamException;
	__objc_class_name__NGCTextStreamLineEnumerator;
	__objc_class_name__NGConcreteFileStreamFileHandle;
	NGStreamSerializeObjC
	NGStreamDeserializeObjC
	NGReadByteFromStream
	NGSafeReadBytesFromStream
	NGSafeWriteBytesToStream
	NGFileReadOnly
	NGFileWriteOnly
	NGFileReadWrite
	NGFileAppend
	NGFileReadAppend
	NGIn
	NGOut
	NGErr
	NGTextIn
	NGTextOut
	NGTextErr
	NGInitStdio
	NGInitTextStdio

;__objc_class_name__NGConcretePipeFileHandle;
;__objc_class_name_NGTaskStream;
