EXPORTS
	__objc_class_name_EOAdaptor;
	__objc_class_name_EOAdaptorChannel;
	__objc_class_name_EOAdaptorContext;
	__objc_class_name_EOArrayProxy;
	__objc_class_name_EOAttribute;
	__objc_class_name_EOAttributeOrdering;
	__objc_class_name_EODatabase;
	__objc_class_name_EODatabaseChannel;
	__objc_class_name_EODatabaseContext;
	__objc_class_name_EODatabaseFault;
	__objc_class_name_EODatabaseFaultResolver;
	__objc_class_name_EOArrayFault;
	__objc_class_name_EOObjectFault;
	__objc_class_name_EOEntity;
	__objc_class_name_EOEntityClassDescription;
	__objc_class_name_EOExpressionArray;
	__objc_class_name_EOFException;
	__objc_class_name_ObjectNotAvailableException;
	__objc_class_name_PropertyDefinitionException;
	__objc_class_name_DestinationEntityDoesntMatchDefinitionException;
	__objc_class_name_InvalidNameException;
	__objc_class_name_InvalidPropertyException;
	__objc_class_name_RelationshipMustBeToOneException;
	__objc_class_name_InvalidValueTypeException;
	__objc_class_name_InvalidAttributeException;
	__objc_class_name_InvalidQualifierException;
	__objc_class_name_EOAdaptorException;
	__objc_class_name_CannotFindAdaptorBundleException;
	__objc_class_name_InvalidAdaptorBundleException;
	__objc_class_name_InvalidAdaptorStateException;
	__objc_class_name_DataTypeMappingNotSupportedException;
	__objc_class_name_ChannelIsNotOpenedException;
	__objc_class_name_AdaptorIsFetchingException;
	__objc_class_name_AdaptorIsNotFetchingException;
	__objc_class_name_NoTransactionInProgressException;
	__objc_class_name_TooManyOpenedChannelsException;
	__objc_class_name_EOKeySortOrdering;
	__objc_class_name_EOModel;
	__objc_class_name_EOModelGroup;
	__objc_class_name_EOObjectUniquer;
	__objc_class_name_EOSinglePrimaryKeyDictionary;
	__objc_class_name_EOSinglePrimaryKeyDictionaryEnumerator;
	__objc_class_name_EOMultiplePrimaryKeyDictionary;
	__objc_class_name_EOPrimaryKeyDictionary;
	__objc_class_name_EOQualifierScannerHandler;
	__objc_class_name_EOQualifierEnumScannerHandler;
	__objc_class_name_EOQuotedExpression;
	__objc_class_name_EORelationship;
	__objc_class_name_EOJoin;
	__objc_class_name_EOSelectScannerHandler;
	__objc_class_name_EOInsertUpdateScannerHandler;
	__objc_class_name_EOSQLExpression;
	__objc_class_name_EOSelectSQLExpression;
	__objc_class_name_EOInsertSQLExpression;
	__objc_class_name_EOUpdateSQLExpression;
	__objc_class_name_EODeleteSQLExpression;
	__objc_class_name_EOQualifierJoinHolder;
	__objc_class_name_EOSQLQualifier;
	__objc_class_name_eoaccess;
