obj/EOAttribute+Oracle.o: EOAttribute+Oracle.m EOAttribute+Oracle.h \
  ../GDLAccess/EOAttribute.h /usr/include/GNUstep/Foundation/NSString.h \
  /usr/include/GNUstep/GNUstepBase/GSVersionMacros.h \
  /usr/include/GNUstep/Foundation/NSObject.h \
  /usr/include/GNUstep/Foundation/NSObjCRuntime.h \
  /usr/include/GNUstep/GNUstepBase/preface.h \
  /usr/include/GNUstep/GNUstepBase/objc-gnu2next.h \
  /usr/include/GNUstep/GNUstepBase/GSConfig.h \
  /usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h \
  /usr/include/GNUstep/Foundation/NSZone.h \
  /usr/include/GNUstep/GNUstepBase/GNUstep.h \
  /usr/include/GNUstep/Foundation/NSDate.h \
  /usr/include/GNUstep/Foundation/NSRange.h \
  /usr/include/GNUstep/Foundation/NSDictionary.h \
  /usr/include/oracle/10.2.0.3/client/oci.h \
  /usr/include/oracle/10.2.0.3/client/oratypes.h \
  /usr/include/oracle/10.2.0.3/client/ocidfn.h \
  /usr/include/oracle/10.2.0.3/client/oci1.h \
  /usr/include/oracle/10.2.0.3/client/oro.h \
  /usr/include/oracle/10.2.0.3/client/ori.h \
  /usr/include/oracle/10.2.0.3/client/ort.h \
  /usr/include/oracle/10.2.0.3/client/orl.h \
  /usr/include/oracle/10.2.0.3/client/ociextp.h \
  /usr/include/oracle/10.2.0.3/client/ociapr.h \
  /usr/include/oracle/10.2.0.3/client/ociap.h \
  /usr/include/oracle/10.2.0.3/client/nzt.h \
  /usr/include/oracle/10.2.0.3/client/nzerror.h \
  /usr/include/oracle/10.2.0.3/client/ocixmldb.h \
  /usr/include/oracle/10.2.0.3/client/oci8dp.h

EOAttribute+Oracle.h:

../GDLAccess/EOAttribute.h:

/usr/include/GNUstep/Foundation/NSString.h:

/usr/include/GNUstep/GNUstepBase/GSVersionMacros.h:

/usr/include/GNUstep/Foundation/NSObject.h:

/usr/include/GNUstep/Foundation/NSObjCRuntime.h:

/usr/include/GNUstep/GNUstepBase/preface.h:

/usr/include/GNUstep/GNUstepBase/objc-gnu2next.h:

/usr/include/GNUstep/GNUstepBase/GSConfig.h:

/usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h:

/usr/include/GNUstep/Foundation/NSZone.h:

/usr/include/GNUstep/GNUstepBase/GNUstep.h:

/usr/include/GNUstep/Foundation/NSDate.h:

/usr/include/GNUstep/Foundation/NSRange.h:

/usr/include/GNUstep/Foundation/NSDictionary.h:

/usr/include/oracle/10.2.0.3/client/oci.h:

/usr/include/oracle/10.2.0.3/client/oratypes.h:

/usr/include/oracle/10.2.0.3/client/ocidfn.h:

/usr/include/oracle/10.2.0.3/client/oci1.h:

/usr/include/oracle/10.2.0.3/client/oro.h:

/usr/include/oracle/10.2.0.3/client/ori.h:

/usr/include/oracle/10.2.0.3/client/ort.h:

/usr/include/oracle/10.2.0.3/client/orl.h:

/usr/include/oracle/10.2.0.3/client/ociextp.h:

/usr/include/oracle/10.2.0.3/client/ociapr.h:

/usr/include/oracle/10.2.0.3/client/ociap.h:

/usr/include/oracle/10.2.0.3/client/nzt.h:

/usr/include/oracle/10.2.0.3/client/nzerror.h:

/usr/include/oracle/10.2.0.3/client/ocixmldb.h:

/usr/include/oracle/10.2.0.3/client/oci8dp.h:
