obj/SaxLocator.o: SaxLocator.m SaxLocator.h \
  /usr/include/GNUstep/Foundation/NSObject.h \
  /usr/include/GNUstep/Foundation/NSObjCRuntime.h \
  /usr/include/GNUstep/GNUstepBase/GSVersionMacros.h \
  /usr/include/GNUstep/GNUstepBase/preface.h \
  /usr/include/GNUstep/GNUstepBase/objc-gnu2next.h \
  /usr/include/GNUstep/GNUstepBase/GSConfig.h \
  /usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h \
  /usr/include/GNUstep/Foundation/NSZone.h \
  /usr/include/GNUstep/GNUstepBase/GNUstep.h \
  /usr/include/GNUstep/Foundation/NSDate.h common.h \
  /usr/include/GNUstep/Foundation/Foundation.h \
  /usr/include/GNUstep/Foundation/FoundationErrors.h \
  /usr/include/GNUstep/Foundation/NSDebug.h \
  /usr/include/GNUstep/Foundation/NSProcessInfo.h \
  /usr/include/GNUstep/Foundation/NSAffineTransform.h \
  /usr/include/GNUstep/Foundation/NSGeometry.h \
  /usr/include/GNUstep/Foundation/NSString.h \
  /usr/include/GNUstep/Foundation/NSRange.h \
  /usr/include/GNUstep/Foundation/NSArchiver.h \
  /usr/include/GNUstep/Foundation/NSCoder.h \
  /usr/include/GNUstep/Foundation/NSArray.h \
  /usr/include/GNUstep/Foundation/NSAttributedString.h \
  /usr/include/GNUstep/Foundation/NSDictionary.h \
  /usr/include/GNUstep/Foundation/NSAutoreleasePool.h \
  /usr/include/GNUstep/Foundation/NSBundle.h \
  /usr/include/GNUstep/Foundation/NSByteOrder.h \
  /usr/include/GNUstep/Foundation/NSCalendarDate.h \
  /usr/include/GNUstep/Foundation/NSCharacterSet.h \
  /usr/include/GNUstep/Foundation/NSClassDescription.h \
  /usr/include/GNUstep/Foundation/NSException.h \
  /usr/include/GNUstep/Foundation/NSComparisonPredicate.h \
  /usr/include/GNUstep/Foundation/NSExpression.h \
  /usr/include/GNUstep/Foundation/NSPredicate.h \
  /usr/include/GNUstep/Foundation/NSCompoundPredicate.h \
  /usr/include/GNUstep/Foundation/NSConnection.h \
  /usr/include/GNUstep/Foundation/NSTimer.h \
  /usr/include/GNUstep/Foundation/NSRunLoop.h \
  /usr/include/GNUstep/Foundation/NSMapTable.h \
  /usr/include/GNUstep/Foundation/NSData.h \
  /usr/include/GNUstep/Foundation/NSSerialization.h \
  /usr/include/GNUstep/Foundation/NSDateFormatter.h \
  /usr/include/GNUstep/Foundation/NSFormatter.h \
  /usr/include/GNUstep/Foundation/NSDecimalNumber.h \
  /usr/include/GNUstep/Foundation/NSDecimal.h \
  /usr/include/GNUstep/Foundation/NSValue.h \
  /usr/include/GNUstep/Foundation/NSDistantObject.h \
  /usr/include/GNUstep/Foundation/NSProxy.h \
  /usr/include/GNUstep/Foundation/NSDistributedLock.h \
  /usr/include/GNUstep/Foundation/NSDistributedNotificationCenter.h \
  /usr/include/GNUstep/Foundation/NSLock.h \
  /usr/include/GNUstep/Foundation/NSNotification.h \
  /usr/include/GNUstep/Foundation/NSEnumerator.h \
  /usr/include/GNUstep/Foundation/NSError.h \
  /usr/include/GNUstep/Foundation/NSFileHandle.h \
  /usr/include/GNUstep/Foundation/NSFileManager.h \
  /usr/include/GNUstep/Foundation/NSHashTable.h \
  /usr/include/GNUstep/Foundation/NSHost.h \
  /usr/include/GNUstep/Foundation/NSHTTPCookie.h \
  /usr/include/GNUstep/Foundation/NSHTTPCookieStorage.h \
  /usr/include/GNUstep/Foundation/NSIndexPath.h \
  /usr/include/GNUstep/Foundation/NSIndexSet.h \
  /usr/include/GNUstep/Foundation/NSInvocation.h \
  /usr/include/GNUstep/Foundation/NSMethodSignature.h \
  /usr/include/GNUstep/Foundation/NSKeyedArchiver.h \
  /usr/include/GNUstep/Foundation/NSPropertyList.h \
  /usr/include/GNUstep/Foundation/NSKeyValueCoding.h \
  /usr/include/GNUstep/Foundation/NSNotificationQueue.h \
  /usr/include/GNUstep/Foundation/NSNetServices.h \
  /usr/include/GNUstep/Foundation/NSNull.h \
  /usr/include/GNUstep/Foundation/NSNumberFormatter.h \
  /usr/include/GNUstep/Foundation/NSPathUtilities.h \
  /usr/include/GNUstep/Foundation/NSPortCoder.h \
  /usr/include/GNUstep/Foundation/NSPortMessage.h \
  /usr/include/GNUstep/Foundation/NSPort.h \
  /usr/include/GNUstep/Foundation/NSPortNameServer.h \
  /usr/include/GNUstep/Foundation/NSProtocolChecker.h \
  /usr/include/GNUstep/Foundation/NSScanner.h \
  /usr/include/GNUstep/Foundation/NSSet.h \
  /usr/include/GNUstep/Foundation/NSSortDescriptor.h \
  /usr/include/GNUstep/Foundation/NSSpellServer.h \
  /usr/include/GNUstep/Foundation/NSStream.h \
  /usr/include/GNUstep/Foundation/NSTask.h \
  /usr/include/GNUstep/Foundation/NSThread.h \
  /usr/include/GNUstep/Foundation/NSTimeZone.h \
  /usr/include/GNUstep/Foundation/NSUndoManager.h \
  /usr/include/GNUstep/Foundation/NSURLAuthenticationChallenge.h \
  /usr/include/GNUstep/Foundation/NSURLCache.h \
  /usr/include/GNUstep/Foundation/NSURLConnection.h \
  /usr/include/GNUstep/Foundation/NSURLCredential.h \
  /usr/include/GNUstep/Foundation/NSURLCredentialStorage.h \
  /usr/include/GNUstep/Foundation/NSURLDownload.h \
  /usr/include/GNUstep/Foundation/NSURLError.h \
  /usr/include/GNUstep/Foundation/NSURL.h \
  /usr/include/GNUstep/Foundation/NSURLHandle.h \
  /usr/include/GNUstep/Foundation/NSURLProtectionSpace.h \
  /usr/include/GNUstep/Foundation/NSURLProtocol.h \
  /usr/include/GNUstep/Foundation/NSURLRequest.h \
  /usr/include/GNUstep/Foundation/NSURLResponse.h \
  /usr/include/GNUstep/Foundation/NSUserDefaults.h \
  /usr/include/GNUstep/Foundation/NSValueTransformer.h \
  /usr/include/GNUstep/Foundation/NSXMLParser.h

SaxLocator.h:

/usr/include/GNUstep/Foundation/NSObject.h:

/usr/include/GNUstep/Foundation/NSObjCRuntime.h:

/usr/include/GNUstep/GNUstepBase/GSVersionMacros.h:

/usr/include/GNUstep/GNUstepBase/preface.h:

/usr/include/GNUstep/GNUstepBase/objc-gnu2next.h:

/usr/include/GNUstep/GNUstepBase/GSConfig.h:

/usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h:

/usr/include/GNUstep/Foundation/NSZone.h:

/usr/include/GNUstep/GNUstepBase/GNUstep.h:

/usr/include/GNUstep/Foundation/NSDate.h:

common.h:

/usr/include/GNUstep/Foundation/Foundation.h:

/usr/include/GNUstep/Foundation/FoundationErrors.h:

/usr/include/GNUstep/Foundation/NSDebug.h:

/usr/include/GNUstep/Foundation/NSProcessInfo.h:

/usr/include/GNUstep/Foundation/NSAffineTransform.h:

/usr/include/GNUstep/Foundation/NSGeometry.h:

/usr/include/GNUstep/Foundation/NSString.h:

/usr/include/GNUstep/Foundation/NSRange.h:

/usr/include/GNUstep/Foundation/NSArchiver.h:

/usr/include/GNUstep/Foundation/NSCoder.h:

/usr/include/GNUstep/Foundation/NSArray.h:

/usr/include/GNUstep/Foundation/NSAttributedString.h:

/usr/include/GNUstep/Foundation/NSDictionary.h:

/usr/include/GNUstep/Foundation/NSAutoreleasePool.h:

/usr/include/GNUstep/Foundation/NSBundle.h:

/usr/include/GNUstep/Foundation/NSByteOrder.h:

/usr/include/GNUstep/Foundation/NSCalendarDate.h:

/usr/include/GNUstep/Foundation/NSCharacterSet.h:

/usr/include/GNUstep/Foundation/NSClassDescription.h:

/usr/include/GNUstep/Foundation/NSException.h:

/usr/include/GNUstep/Foundation/NSComparisonPredicate.h:

/usr/include/GNUstep/Foundation/NSExpression.h:

/usr/include/GNUstep/Foundation/NSPredicate.h:

/usr/include/GNUstep/Foundation/NSCompoundPredicate.h:

/usr/include/GNUstep/Foundation/NSConnection.h:

/usr/include/GNUstep/Foundation/NSTimer.h:

/usr/include/GNUstep/Foundation/NSRunLoop.h:

/usr/include/GNUstep/Foundation/NSMapTable.h:

/usr/include/GNUstep/Foundation/NSData.h:

/usr/include/GNUstep/Foundation/NSSerialization.h:

/usr/include/GNUstep/Foundation/NSDateFormatter.h:

/usr/include/GNUstep/Foundation/NSFormatter.h:

/usr/include/GNUstep/Foundation/NSDecimalNumber.h:

/usr/include/GNUstep/Foundation/NSDecimal.h:

/usr/include/GNUstep/Foundation/NSValue.h:

/usr/include/GNUstep/Foundation/NSDistantObject.h:

/usr/include/GNUstep/Foundation/NSProxy.h:

/usr/include/GNUstep/Foundation/NSDistributedLock.h:

/usr/include/GNUstep/Foundation/NSDistributedNotificationCenter.h:

/usr/include/GNUstep/Foundation/NSLock.h:

/usr/include/GNUstep/Foundation/NSNotification.h:

/usr/include/GNUstep/Foundation/NSEnumerator.h:

/usr/include/GNUstep/Foundation/NSError.h:

/usr/include/GNUstep/Foundation/NSFileHandle.h:

/usr/include/GNUstep/Foundation/NSFileManager.h:

/usr/include/GNUstep/Foundation/NSHashTable.h:

/usr/include/GNUstep/Foundation/NSHost.h:

/usr/include/GNUstep/Foundation/NSHTTPCookie.h:

/usr/include/GNUstep/Foundation/NSHTTPCookieStorage.h:

/usr/include/GNUstep/Foundation/NSIndexPath.h:

/usr/include/GNUstep/Foundation/NSIndexSet.h:

/usr/include/GNUstep/Foundation/NSInvocation.h:

/usr/include/GNUstep/Foundation/NSMethodSignature.h:

/usr/include/GNUstep/Foundation/NSKeyedArchiver.h:

/usr/include/GNUstep/Foundation/NSPropertyList.h:

/usr/include/GNUstep/Foundation/NSKeyValueCoding.h:

/usr/include/GNUstep/Foundation/NSNotificationQueue.h:

/usr/include/GNUstep/Foundation/NSNetServices.h:

/usr/include/GNUstep/Foundation/NSNull.h:

/usr/include/GNUstep/Foundation/NSNumberFormatter.h:

/usr/include/GNUstep/Foundation/NSPathUtilities.h:

/usr/include/GNUstep/Foundation/NSPortCoder.h:

/usr/include/GNUstep/Foundation/NSPortMessage.h:

/usr/include/GNUstep/Foundation/NSPort.h:

/usr/include/GNUstep/Foundation/NSPortNameServer.h:

/usr/include/GNUstep/Foundation/NSProtocolChecker.h:

/usr/include/GNUstep/Foundation/NSScanner.h:

/usr/include/GNUstep/Foundation/NSSet.h:

/usr/include/GNUstep/Foundation/NSSortDescriptor.h:

/usr/include/GNUstep/Foundation/NSSpellServer.h:

/usr/include/GNUstep/Foundation/NSStream.h:

/usr/include/GNUstep/Foundation/NSTask.h:

/usr/include/GNUstep/Foundation/NSThread.h:

/usr/include/GNUstep/Foundation/NSTimeZone.h:

/usr/include/GNUstep/Foundation/NSUndoManager.h:

/usr/include/GNUstep/Foundation/NSURLAuthenticationChallenge.h:

/usr/include/GNUstep/Foundation/NSURLCache.h:

/usr/include/GNUstep/Foundation/NSURLConnection.h:

/usr/include/GNUstep/Foundation/NSURLCredential.h:

/usr/include/GNUstep/Foundation/NSURLCredentialStorage.h:

/usr/include/GNUstep/Foundation/NSURLDownload.h:

/usr/include/GNUstep/Foundation/NSURLError.h:

/usr/include/GNUstep/Foundation/NSURL.h:

/usr/include/GNUstep/Foundation/NSURLHandle.h:

/usr/include/GNUstep/Foundation/NSURLProtectionSpace.h:

/usr/include/GNUstep/Foundation/NSURLProtocol.h:

/usr/include/GNUstep/Foundation/NSURLRequest.h:

/usr/include/GNUstep/Foundation/NSURLResponse.h:

/usr/include/GNUstep/Foundation/NSUserDefaults.h:

/usr/include/GNUstep/Foundation/NSValueTransformer.h:

/usr/include/GNUstep/Foundation/NSXMLParser.h:
