/* 
   EOKeyGlobalID+PGVal.m

   Copyright (C) 2006 SKYRIX Software AG and Helge Hess

   Author: Helge Hess (helge.hess@opengroupware.org)

   This file is part of the PostgreSQL Adaptor Library

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <EOControl/EOKeyGlobalID.h>
#include "PostgreSQL72Channel.h"
#include "common.h"

@implementation EOKeyGlobalID(PostgreSQL72Values)

- (NSString *)stringValueForPostgreSQLType:(NSString *)_type
  attribute:(EOAttribute *)_attribute
{
  if (self->count == 0) {
    NSLog(@"ERROR(%s): got a EOKeyGlobalID w/o key values: %@",
	  __PRETTY_FUNCTION__, self);
    return nil;
  }
  
  return [self->values[0] stringValueForPostgreSQLType:_type 
	                  attribute:_attribute];
}

@end /* EOKeyGlobalID(PostgreSQL72Values) */
